# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import numpy as np
import pytest

from power_grid_model.core.data_handling import OutputType, create_output_data, process_output_component_types
from power_grid_model.core.dataset_definitions import ComponentType as CT, DatasetType as DT
from power_grid_model.core.power_grid_meta import initialize_array


@pytest.mark.parametrize(
    ("output_component_types", "is_batch", "expected"),
    [
        (
            None,
            False,
            {
                CT.node: initialize_array(DT.sym_output, CT.node, 4),
                CT.sym_load: initialize_array(DT.sym_output, CT.sym_load, 3),
                CT.source: initialize_array(DT.sym_output, CT.source, 1),
            },
        ),
        (
            [CT.node, CT.sym_load],
            False,
            {
                CT.node: initialize_array(DT.sym_output, CT.node, 4),
                CT.sym_load: initialize_array(DT.sym_output, CT.sym_load, 3),
            },
        ),
        (
            {CT.node, CT.sym_load},
            False,
            {
                CT.node: initialize_array(DT.sym_output, CT.node, 4),
                CT.sym_load: initialize_array(DT.sym_output, CT.sym_load, 3),
            },
        ),
        pytest.param(
            {CT.node: [], CT.sym_load: []}, True, {CT.node: dict(), CT.sym_load: dict()}, marks=pytest.mark.xfail
        ),
        pytest.param({CT.node: [], CT.sym_load: ["p"]}, True, {}, marks=pytest.mark.xfail),
        pytest.param({CT.node: ["u"], CT.sym_load: ["p"]}, True, {}, marks=pytest.mark.xfail),
        pytest.param({CT.node: None, CT.sym_load: ["p"]}, True, {}, marks=pytest.mark.xfail),
    ],
)
def test_create_output_data(output_component_types, is_batch, expected):
    # TODO use is_batch and shorten parameterization after columnar data implementation
    all_component_count = {CT.node: 4, CT.sym_load: 3, CT.source: 1}
    batch_size = 15 if is_batch else 1
    actual = create_output_data(
        output_component_types=output_component_types,
        output_type=OutputType.SYM_OUTPUT,
        all_component_count=all_component_count,
        is_batch=is_batch,
        batch_size=batch_size,
    )

    assert actual.keys() == expected.keys()
    for comp in expected:
        if isinstance(expected[comp], np.ndarray):
            # Row based
            assert actual[comp].dtype == expected[comp].dtype
        elif expected[comp] == dict():
            # Empty atrtibutes
            assert actual[comp] == expected[comp]
        else:
            # Columnar data
            assert actual[comp].keys() == expected[comp].keys()
            assert all(actual[comp][attr].dtype == expected[comp][attr].dtype for attr in expected[comp])


@pytest.mark.parametrize(
    ("output_component_types", "error", "match"),
    [
        ({"abc": 3, "def": None}, ValueError, "Invalid output_component_types"),
        ({"abc": None, "def": None}, KeyError, "unknown component"),
        ({"abc": None, CT.sym_load: None}, KeyError, "unknown component"),
        ({"abc": ["xyz"], CT.sym_load: None}, KeyError, "unknown component"),
        ({CT.node: ["xyz"], CT.sym_load: None}, KeyError, "unknown attributes"),
        ({CT.node: ["xyz1"], CT.sym_load: ["xyz2"]}, KeyError, "unknown attributes"),
    ],
)
def test_create_output_data__errors(output_component_types, error, match):
    available_components = [CT.node, CT.sym_load, CT.source]
    with pytest.raises(error, match=match):
        process_output_component_types(
            output_type=OutputType.SYM_OUTPUT,
            output_component_types=output_component_types,
            available_components=available_components,
        )



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Type hints for PGM. This includes all miscellaneous type hints not under dataset or dataset_definitions categories
"""
from power_grid_model.core.dataset_definitions import ComponentType

_ComponentAttributeMappingDict = dict[ComponentType, set[str] | list[str] | None]

ComponentAttributeMapping = set[ComponentType] | list[ComponentType] | None | _ComponentAttributeMappingDict



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Error classes
"""


import numpy as np


class PowerGridError(RuntimeError):
    """Generic power grid error."""


class PowerGridBatchError(PowerGridError):
    """Error occurs in batch calculation."""

    failed_scenarios: np.ndarray
    succeeded_scenarios: np.ndarray
    error_messages: list[str]
    errors: list[PowerGridError]


class InvalidArguments(PowerGridError):
    """A (combination of) input arguments is not valid."""


class MissingCaseForEnumError(InvalidArguments):
    """An enum value is not covered in a for loop.

    This usually happens when an invalid combination of (enum) settings is provided."""


class ConflictVoltage(PowerGridError):
    """There is a confliciting voltage"""


class InvalidBranch(PowerGridError):
    """A branch is invalid."""


class InvalidBranch3(PowerGridError):
    """A branch3 is invalid."""


class InvalidTransformerClock(PowerGridError):
    """Invalid transformer clock found."""


class SparseMatrixError(PowerGridError):
    """Attempting to invert a non-invertible matrix."""


class NotObservableError(SparseMatrixError):
    """Attempting to solve a non-observable system."""


class IterationDiverge(PowerGridError):
    """Unable to iteratively converge to an optimum within the set number of iterations and precision."""


class MaxIterationReached(IterationDiverge):
    """Maximum number of iterations reached."""


class InvalidID(PowerGridError):
    """An ID is invalid."""


class ConflictID(InvalidID):
    """Conflicting IDs found."""


class IDNotFound(InvalidID):
    """A reference to a non-existent ID was provided."""


class InvalidMeasuredObject(InvalidID):
    """A provided measured object is invalid."""


class InvalidRegulatedObject(InvalidID):
    """A provided regulated object is invalid."""


class IDWrongType(InvalidID):
    """A referenced ID points to a component that cannot be referenced here."""


class InvalidCalculationMethod(PowerGridError):
    """Invalid calculation method provided."""


class AutomaticTapCalculationError(PowerGridError):
    """Automatic tap changer with tap at LV side is unsupported for automatic tap changing calculation."""


class InvalidShortCircuitPhaseOrType(PowerGridError):
    """Invalid (combination of) short circuit types and phase(s) provided."""


class PowerGridSerializationError(PowerGridError):
    """Error occurs during (de-)serialization."""


class PowerGridDatasetError(PowerGridError):
    """Error occurs during dataset handling."""


class PowerGridUnreachableHitError(PowerGridError):
    """Supposedly unreachable code was hit.

    This usually means a failed assumption and may be caused by a bug in the PGM library."""



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Data types for power grid model dataset and component types."""

# This file is automatically generated. DO NOT modify it manually!

from enum import Enum, EnumMeta
from typing import Any, Mapping

# pylint: disable=invalid-name

# fmt: off

class _MetaEnum(EnumMeta):
    def __contains__(cls, member):
        """
        Check if member is part of the Enum.

        Args:
            member: Member to check.

        Returns:
            bool: True if the member is part of the Enum, False otherwise.
        """
        return member in cls.__members__.keys()


class DatasetType(str, Enum, metaclass=_MetaEnum):
    """
    A DatasetType is the type of a :class:`Dataset` in power grid model.

    - Examples:

        - DatasetType.input = "input"
        - DatasetType.update = "update"
    """

    input = "input"
    sym_output = "sym_output"
    asym_output = "asym_output"
    update = "update"
    sc_output = "sc_output"


class ComponentType(str, Enum, metaclass=_MetaEnum):
    """
    A ComponentType is the type of a grid component.

    - Examples:

        - ComponentType.node = "node"
        - ComponentType.line = "line"
    """

    node = "node"
    line = "line"
    link = "link"
    transformer = "transformer"
    transformer_tap_regulator = "transformer_tap_regulator"
    three_winding_transformer = "three_winding_transformer"
    sym_load = "sym_load"
    sym_gen = "sym_gen"
    asym_load = "asym_load"
    asym_gen = "asym_gen"
    shunt = "shunt"
    source = "source"
    sym_voltage_sensor = "sym_voltage_sensor"
    asym_voltage_sensor = "asym_voltage_sensor"
    sym_power_sensor = "sym_power_sensor"
    asym_power_sensor = "asym_power_sensor"
    fault = "fault"


# pylint: enable=invalid-name


def _str_to_datatype(data_type: Any) -> DatasetType:
    """Helper function to transform data_type str to DatasetType."""
    if isinstance(data_type, str):
        return DatasetType[data_type]
    return data_type


def _map_to_datatypes(data: Mapping[Any, Any]) -> dict[DatasetType, Any]:
    """Helper function to map datatype str keys to DatasetType."""
    return {_str_to_datatype(key): value for key, value in data.items()}


def _str_to_component_type(component: Any) -> ComponentType:
    """Helper function to transform component str to ComponentType."""
    if isinstance(component, str):
        return ComponentType[component]
    return component


def _map_to_component_types(data: Mapping[Any, Any]) -> dict[ComponentType, Any]:
    """Helper function to map componenttype str keys to ComponentType."""
    return {_str_to_component_type(key): value for key, value in data.items()}


# fmt: on



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Data handling
"""


from enum import Enum
from typing import Mapping

import numpy as np

from power_grid_model.core.dataset_definitions import ComponentType, DatasetType
from power_grid_model.core.power_grid_dataset import CConstDataset, CMutableDataset
from power_grid_model.core.power_grid_meta import initialize_array, power_grid_meta_data
from power_grid_model.data_types import Dataset
from power_grid_model.enum import CalculationType
from power_grid_model.typing import ComponentAttributeMapping, _ComponentAttributeMappingDict


class OutputType(Enum):
    """
    The different supported output types:
        - DatasetType.sym_output
        - DatasetType.asym_output
        - DatasetType.sc_output
    """

    SYM_OUTPUT = DatasetType.sym_output
    ASYM_OUTPUT = DatasetType.asym_output
    SC_OUTPUT = DatasetType.sc_output


def get_output_type(*, calculation_type: CalculationType, symmetric: bool) -> OutputType:
    """
    Get the output type based on the provided arguments.

    Args:
        calculation_type:
            request the output type for a specific calculation type (power_flow, state_estimation, ...)
        symmetric:
            True: three-phase symmetric calculation, even for asymmetric loads/generations
            False: three-phase asymmetric calculation

    Returns:
        the output type that fits the format requested by the output type
    """
    if calculation_type in (CalculationType.power_flow, CalculationType.state_estimation):
        return OutputType.SYM_OUTPUT if symmetric else OutputType.ASYM_OUTPUT

    if calculation_type == CalculationType.short_circuit:
        return OutputType.SC_OUTPUT

    raise NotImplementedError()


def prepare_input_view(input_data: Mapping[ComponentType, np.ndarray]) -> CConstDataset:
    """
    Create a view of the input data in a format compatible with the PGM core libary.

    Args:
        input_data:
            the input data to create the view from

    Returns:
        instance of CConstDataset ready to be fed into C API
    """
    return CConstDataset(input_data, dataset_type=DatasetType.input)


def prepare_update_view(update_data: Mapping[ComponentType, np.ndarray | Mapping[str, np.ndarray]]) -> CConstDataset:
    """
    Create a view of the update data, or an empty view if not provided, in a format compatible with the PGM core libary.

    Args:
        update_data:
            the update data to create the view from. Defaults to None

    Returns:
        instance of CConstDataset ready to be fed into C API
    """
    return CConstDataset(update_data, dataset_type=DatasetType.update)


def prepare_output_view(output_data: Mapping[ComponentType, np.ndarray], output_type: OutputType) -> CMutableDataset:
    """
    create a view of the output data in a format compatible with the PGM core libary.

    Args:
        output_data:
            the output data to create the view from
        output_type:
            the output type of the output_data

    Returns:
        instance of CMutableDataset ready to be fed into C API
    """
    return CMutableDataset(output_data, dataset_type=output_type.value)


def create_output_data(
    output_component_types: ComponentAttributeMapping,
    output_type: OutputType,
    all_component_count: dict[ComponentType, int],
    is_batch: bool,
    batch_size: int,
) -> Dataset:
    """
    Create the output dataset based on component and batch size from the model; and output attributes requested by user.

    Args:
        output_component_types:
            the output components the user seeks to extract
        output_type:
            the type of output that the user will see (as per the calculation options)
        all_component_count:
            the amount of components in the grid (as per the input data)
        is_batch:
            if the dataset is batch
        batch_size:
            the batch size

    Returns:
        Dataset: output dataset
    """
    processed_output_types = process_output_component_types(
        output_type, output_component_types, list(all_component_count.keys())
    )

    all_component_count = {k: v for k, v in all_component_count.items() if k in processed_output_types}

    # create result dataset
    result_dict = {}

    for name, count in all_component_count.items():
        # shape
        if is_batch:
            shape: tuple[int] | tuple[int, int] = (batch_size, count)
        else:
            shape = (count,)
        result_dict[name] = initialize_array(output_type.value, name, shape=shape, empty=True)
    return result_dict


def process_output_component_types(
    output_type: OutputType,
    output_component_types: ComponentAttributeMapping,
    available_components: list[ComponentType],
) -> _ComponentAttributeMappingDict:
    """Checks valid type for output_component_types. Also checks for any invalid component names and attribute names

    Args:
        output_type (OutputType): the type of output that the user will see (as per the calculation options)
        output_component_types (OutputComponentNamesType):  output_component_types provided by user
        available_components (list[ComponentType]):  all components available in model instance

    Raises:
        ValueError: when the type for output_comoponent_types is incorrect
        KeyError: with "unknown component" for any unknown components
        KeyError: with "unknown attributes" for any unknown attributes for a known component

    Returns:
        _OutputComponentTypeDict: processed output_component_types in a dictionary
    """
    # limit all component count to user specified component types in output and convert to a dict
    if output_component_types is None:
        output_component_types = {k: None for k in available_components}
    elif isinstance(output_component_types, (list, set)):
        output_component_types = {k: None for k in output_component_types}
    elif not isinstance(output_component_types, dict) or not all(
        attrs is None or isinstance(attrs, (set, list)) for attrs in output_component_types.values()
    ):
        raise ValueError(f"Invalid output_component_types provided: {output_component_types}")

    # raise error if some specified components are unknown
    output_meta = power_grid_meta_data[output_type.value]
    unknown_components = [x for x in output_component_types if x not in output_meta]
    if unknown_components:
        raise KeyError(f"You have specified some unknown component types: {unknown_components}")

    unknown_attributes = {}
    for comp_name, attrs in output_component_types.items():
        if attrs is None:
            continue
        diff = set(attrs).difference(output_meta[comp_name].dtype.names)
        if diff != set():
            unknown_attributes[comp_name] = diff

    if unknown_attributes:
        raise KeyError(f"You have specified some unknown attributes: {unknown_attributes}")

    return output_component_types



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from pathlib import Path
from unittest.mock import MagicMock, mock_open, patch

import numpy as np
import pytest

from power_grid_model import LoadGenType, initialize_array
from power_grid_model.core.power_grid_meta import power_grid_meta_data
from power_grid_model.data_types import Dataset
from power_grid_model.utils import (
    export_json_data,
    get_component_batch_size,
    get_dataset_batch_size,
    get_dataset_scenario,
    json_deserialize_from_file,
    json_serialize_to_file,
    msgpack_deserialize_from_file,
    msgpack_serialize_to_file,
    self_test,
)


def test_get_dataset_scenario():
    data = {
        "foo": np.array([["bar", "baz"], ["foobar", "foobaz"]]),
        "hi": {
            "data": np.array(["hello", "hey"]),
            "indptr": np.array([0, 0, 2]),
        },
    }
    result = get_dataset_scenario(data, 0)
    assert result.keys() == data.keys()
    np.testing.assert_array_equal(result["foo"], data["foo"][0])
    np.testing.assert_array_equal(result["hi"], data["hi"]["data"][0:0])

    result = get_dataset_scenario(data, 1)
    assert result.keys() == data.keys()
    np.testing.assert_array_equal(result["foo"], data["foo"][1])
    np.testing.assert_array_equal(result["hi"], data["hi"]["data"][0:2])

    with pytest.raises(IndexError):
        get_dataset_scenario(data, 2)


def test_get_data_set_batch_size():
    line = initialize_array("update", "line", (3, 2))
    line["id"] = [[5, 6], [6, 7], [7, 5]]
    line["from_status"] = [[1, 1], [1, 1], [1, 1]]

    asym_load = initialize_array("update", "asym_load", (3, 2))
    asym_load["id"] = [[9, 10], [9, 10], [9, 10]]

    batch_data = {"line": line, "asym_load": asym_load}

    assert get_dataset_batch_size(batch_data) == 3


def test_get_dataset_batch_size_sparse():
    data = {
        "node": {
            "data": np.zeros(shape=3, dtype=power_grid_meta_data["input"]["node"]),
            "indptr": np.array([0, 2, 3, 3]),
        },
        "sym_load": {
            "data": np.zeros(shape=2, dtype=power_grid_meta_data["input"]["sym_load"]),
            "indptr": np.array([0, 0, 1, 2]),
        },
        "asym_load": {
            "data": np.zeros(shape=4, dtype=power_grid_meta_data["input"]["asym_load"]),
            "indptr": np.array([0, 2, 3, 4]),
        },
    }

    assert get_dataset_batch_size(data) == 3


def test_get_dataset_batch_size_mixed():
    line = initialize_array("update", "line", (3, 2))
    line["id"] = [[5, 6], [6, 7], [7, 5]]
    line["from_status"] = [[1, 1], [1, 1], [1, 1]]

    asym_load = initialize_array("update", "asym_load", (2, 2))
    asym_load["id"] = [[9, 10], [9, 10]]

    data_dense = {"line": line, "asym_load": asym_load}
    data_sparse = {
        "node": {
            "data": np.zeros(shape=3, dtype=power_grid_meta_data["input"]["node"]),
            "indptr": np.array([0, 2, 3, 3, 5]),
        },
        "sym_load": {
            "data": np.zeros(shape=2, dtype=power_grid_meta_data["input"]["sym_load"]),
            "indptr": np.array([0, 0, 1, 2]),
        },
        "asym_load": {
            "data": np.zeros(shape=4, dtype=power_grid_meta_data["input"]["asym_load"]),
            "indptr": np.array([0, 2, 3]),
        },
    }
    with pytest.raises(ValueError):
        get_dataset_batch_size(data_dense)
    with pytest.raises(ValueError):
        get_dataset_batch_size(data_sparse)


def test_get_component_batch_size():
    asym_load = initialize_array("update", "asym_load", (3, 2))
    asym_load["id"] = [[9, 10], [9, 10], [9, 10]]

    sym_load = {
        "data": np.zeros(shape=2, dtype=power_grid_meta_data["input"]["sym_load"]),
        "indptr": np.array([0, 0, 1, 2]),
    }
    assert get_component_batch_size(asym_load) == 3
    assert get_component_batch_size(sym_load) == 3


@patch("builtins.open", new_callable=mock_open)
@patch("power_grid_model.utils.json_deserialize")
def test_json_deserialize_from_file(deserialize_mock: MagicMock, open_mock: MagicMock):
    handle = open_mock()
    handle.read.return_value = '{"version": "1.0", "data": {"foo": [{"val": 123}]}, "bar": {"baz": 456}}'
    deserialize_mock.return_value = {"foo": [{"val": 123}]}  # type: ignore
    assert json_deserialize_from_file(file_path=Path("output.json")) == deserialize_mock.return_value
    handle.read.assert_called_once()
    deserialize_mock.assert_called_once_with(handle.read.return_value)


@patch("builtins.open", new_callable=mock_open)
@patch("power_grid_model.utils.json_serialize")
def test_json_serialize(serialize_mock: MagicMock, open_mock: MagicMock):
    serialize_mock.return_value = '{"version": "1.0", "data": {"foo": [{"val": 123}]}, "bar": {"baz": 456}}'
    data: Dataset = {}  # type: ignore
    json_serialize_to_file(file_path=Path("output.json"), data=data, use_compact_list=False, indent=2)
    serialize_mock.assert_called_once_with(data=data, dataset_type=None, use_compact_list=False, indent=2)
    handle = open_mock()
    handle.write.assert_called_once_with(serialize_mock.return_value)


@patch("builtins.open", new_callable=mock_open)
@patch("power_grid_model.utils.msgpack_deserialize")
def test_msgpack_deserialize_from_file(deserialize_mock: MagicMock, open_mock: MagicMock):
    handle = open_mock()
    handle.read.return_value = b'{"version": "1.0", "data": {"foo": [{"val": 123}]}, "bar": {"baz": 456}}'
    deserialize_mock.return_value = {"foo": [{"val": 123}]}  # type: ignore
    assert msgpack_deserialize_from_file(file_path=Path("output.msgpack")) == deserialize_mock.return_value
    handle.read.assert_called_once()
    deserialize_mock.assert_called_once_with(handle.read.return_value)


@patch("builtins.open", new_callable=mock_open)
@patch("power_grid_model.utils.msgpack_serialize")
def test_msgpack_serialize(serialize_mock: MagicMock, open_mock: MagicMock):
    serialize_mock.return_value = b'{"version": "1.0", "data": {"foo": [{"val": 123}]}, "bar": {"baz": 456}}'
    data: Dataset = {}  # type: ignore
    msgpack_serialize_to_file(file_path=Path("output.msgpack"), data=data, use_compact_list=False)
    serialize_mock.assert_called_once_with(data=data, dataset_type=None, use_compact_list=False)
    handle = open_mock()
    handle.write.assert_called_once_with(serialize_mock.return_value)


def test_self_test():
    self_test()



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Load meta data from C core and define numpy structured array
"""

from dataclasses import dataclass
from enum import IntEnum
from typing import Any

import numpy as np

from power_grid_model.core.dataset_definitions import (
    ComponentType,
    DatasetType,
    _str_to_component_type,
    _str_to_datatype,
)
from power_grid_model.core.power_grid_core import AttributePtr, ComponentPtr, DatasetPtr, power_grid_core as pgc


# constant enum for ctype
# pylint: disable=invalid-name
class PGMCType(IntEnum):
    """enumeration for ctype"""

    int32 = 0
    int8 = 1
    double = 2
    double3 = 3


_CTYPE_NUMPY_MAP = {PGMCType.double: "f8", PGMCType.int32: "i4", PGMCType.int8: "i1", PGMCType.double3: "(3,)f8"}
_ENDIANNESS = "<" if pgc.is_little_endian() == 1 else ">"
_NAN_VALUE_MAP = {
    f"{_ENDIANNESS}f8": np.nan,
    f"{_ENDIANNESS}(3,)f8": np.nan,
    f"{_ENDIANNESS}i4": np.iinfo(f"{_ENDIANNESS}i4").min,
    f"{_ENDIANNESS}i1": np.iinfo(f"{_ENDIANNESS}i1").min,
}


@dataclass
class ComponentMetaData:
    """
    Data class for component metadata
    """

    dtype: np.dtype
    dtype_dict: dict[str, Any]
    nans: dict[str, float | int]
    nan_scalar: np.ndarray

    def __getitem__(self, item):
        """
        Get item of dataclass

        Args:
            item: item name

        Returns:

        """
        return getattr(self, item)


DatasetMetaData = dict[ComponentType, ComponentMetaData]
PowerGridMetaData = dict[DatasetType, DatasetMetaData]


def _generate_meta_data() -> PowerGridMetaData:
    """

    Returns: meta data for all dataset

    """
    py_meta_data = {}
    n_datasets = pgc.meta_n_datasets()
    for i in range(n_datasets):
        dataset = pgc.meta_get_dataset_by_idx(i)
        py_meta_data[_str_to_datatype(pgc.meta_dataset_name(dataset))] = _generate_meta_dataset(dataset)
    return py_meta_data


def _generate_meta_dataset(dataset: DatasetPtr) -> DatasetMetaData:
    """

    Args:
        dataset: dataset

    Returns: meta data for one dataset

    """
    py_meta_dataset = {}
    n_components = pgc.meta_n_components(dataset)
    for i in range(n_components):
        component = pgc.meta_get_component_by_idx(dataset, i)
        py_meta_dataset[_str_to_component_type(pgc.meta_component_name(component))] = _generate_meta_component(
            component
        )
    return py_meta_dataset


def _generate_meta_component(component: ComponentPtr) -> ComponentMetaData:
    """

    Args:
        component: component

    Returns: meta data for one component

    """

    dtype_dict = _generate_meta_attributes(component)
    dtype = np.dtype({k: v for k, v in dtype_dict.items() if k != "nans"})  # type: ignore
    nans = dict(zip(dtype_dict["names"], dtype_dict["nans"]))
    if dtype.alignment != pgc.meta_component_alignment(component):
        raise TypeError(f'Aligment mismatch for component type: "{pgc.meta_component_name(component)}" !')
    # get single nan scalar
    nan_scalar = np.empty(1, dtype=dtype)
    for key, value in nans.items():
        nan_scalar[key] = value
    return ComponentMetaData(dtype=dtype, dtype_dict=dtype_dict, nans=nans, nan_scalar=nan_scalar)


def _generate_meta_attributes(component: ComponentPtr) -> dict:
    """

    Args:
        component: component

    Returns: meta data for all attributes

    """
    names = []
    formats = []
    offsets = []
    nans = []
    n_attrs = pgc.meta_n_attributes(component)
    for i in range(n_attrs):
        attribute: AttributePtr = pgc.meta_get_attribute_by_idx(component, i)
        attr_name: str = pgc.meta_attribute_name(attribute)
        attr_ctype: int = pgc.meta_attribute_ctype(attribute)
        attr_offset: int = pgc.meta_attribute_offset(attribute)
        attr_np_type = f"{_ENDIANNESS}{_CTYPE_NUMPY_MAP[PGMCType(attr_ctype)]}"
        attr_nan = _NAN_VALUE_MAP[attr_np_type]
        names.append(attr_name)
        formats.append(attr_np_type)
        offsets.append(attr_offset)
        nans.append(attr_nan)
    return {
        "names": names,
        "formats": formats,
        "offsets": offsets,
        "itemsize": pgc.meta_component_size(component),
        "aligned": True,
        "nans": nans,
    }


# store meta data
power_grid_meta_data = _generate_meta_data()


def initialize_array(
    data_type: str | DatasetType,
    component_type: str | ComponentType,
    shape: tuple | int,
    empty: bool = False,
) -> np.ndarray:
    """
    Initializes an array for use in Power Grid Model calculations

    Args:
        data_type: input, update, sym_output, or asym_output
        component_type: one component type, e.g. node
        shape: shape of initialization
            integer, it is a 1-dimensional array
            tuple, it is an N-dimensional (tuple.shape) array
        empty: if True, leave the memory block un-initialized

    Returns:
        np structured array with all entries as null value
    """
    data_type = _str_to_datatype(data_type)
    component_type = _str_to_component_type(component_type)
    if not isinstance(shape, tuple):
        shape = (shape,)
    if empty:
        return np.empty(shape=shape, dtype=power_grid_meta_data[data_type][component_type].dtype, order="C")
    return np.full(
        shape=shape,
        fill_value=power_grid_meta_data[data_type][component_type].nan_scalar,
        dtype=power_grid_meta_data[data_type][component_type].dtype,
        order="C",
    )



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Power Grid Model input/update data validation"""

from power_grid_model.validation.assertions import ValidationException, assert_valid_batch_data, assert_valid_input_data
from power_grid_model.validation.errors import ValidationError
from power_grid_model.validation.utils import errors_to_string
from power_grid_model.validation.validation import validate_batch_data, validate_input_data
